/*
 * Copyright 2004-2020 Sandboxie Holdings, LLC 
 *
 * This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

//---------------------------------------------------------------------------
// Pool Memory Manager
//---------------------------------------------------------------------------


#ifndef _MY_POOL_H
#define _MY_POOL_H


//---------------------------------------------------------------------------
// Structures
//---------------------------------------------------------------------------


typedef struct POOL POOL;


//---------------------------------------------------------------------------
// Functions
//---------------------------------------------------------------------------


#ifdef __cplusplus
extern "C" {
#endif


POOL *Pool_Create(void);
POOL *Pool_CreateTagged(ULONG tag);
ULONG Pool_Delete(POOL *pool);

void *Pool_Alloc(POOL *pool, ULONG size);
void Pool_Free(void *ptr, ULONG size);


#ifdef __cplusplus
} // extern "C"
#endif


//---------------------------------------------------------------------------


#endif // _MY_POOL_H
